/*
Copyright 2024 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package slowset

import (
	"testing"
	"time"
)

func TestSlowSet(t *testing.T) {
	tests := []struct {
		name          string
		retentionTime time.Duration
		resyncPeriod  time.Duration
		testFunc      func(*SlowSet) bool
	}{
		{
			name:         "Should not change time of a key if added multiple times",
			resyncPeriod: 100 * time.Millisecond,
			testFunc: func(s *SlowSet) bool {
				key := "key"
				info := ObjectData{
					Timestamp: time.Now(),
				}
				s.Add(key, info)
				time1 := s.workSet[key]
				s.Add(key, info)
				time2 := s.workSet[key]
				return time1 == time2
			},
		},
		{
			name:          "Should remove key after retention time",
			retentionTime: 200 * time.Millisecond,
			resyncPeriod:  100 * time.Millisecond,
			testFunc: func(s *SlowSet) bool {
				key := "key"
				info := ObjectData{
					Timestamp: time.Now(),
				}
				s.Add(key, info)
				time.Sleep(300 * time.Millisecond)
				return !s.Contains(key)
			},
		},
		{
			name:          "Should not remove key before retention time",
			retentionTime: 200 * time.Millisecond,
			resyncPeriod:  100 * time.Millisecond,
			testFunc: func(s *SlowSet) bool {
				key := "key"
				info := ObjectData{
					Timestamp: time.Now(),
				}
				s.Add(key, info)
				time.Sleep(100 * time.Millisecond)
				return s.Contains(key)
			},
		},
		{
			name:          "Should return time remaining for added keys",
			retentionTime: 300 * time.Millisecond,
			resyncPeriod:  100 * time.Millisecond,
			testFunc: func(s *SlowSet) bool {
				key := "key"
				info := ObjectData{
					Timestamp: time.Now(),
				}
				s.Add(key, info)
				time.Sleep(100 * time.Millisecond)
				timeRemaining := s.TimeRemaining(key)
				return timeRemaining > 0 && timeRemaining < 300*time.Millisecond
			},
		},
		{
			name:          "should return false for Contains if key is present but expired",
			resyncPeriod:  200 * time.Millisecond,
			retentionTime: 300 * time.Millisecond,
			testFunc: func(s *SlowSet) bool {
				key := "key"
				info := ObjectData{
					Timestamp: time.Now(),
				}
				s.Add(key, info)
				time.Sleep(301 * time.Millisecond)
				return !s.Contains(key)
			},
		},
	}

	for i := range tests {
		test := tests[i]
		t.Run(test.name, func(t *testing.T) {
			s := NewSlowSet(test.retentionTime)
			s.resyncPeriod = test.resyncPeriod
			stopCh := make(chan struct{}, 1)
			go s.Run(stopCh)
			defer close(stopCh)
			if !test.testFunc(s) {
				t.Errorf("Test failed")
			}
		})
	}
}
